MODULE Experiment;

IMPORT ML := MutualLigationContT, Kernel, Out;

VAR time: LONGINT;

PROCEDURE Run*;
BEGIN
	ML.Lbd := 1;
	ML.Lad := 7;
	ML.Lbe := 7;
	ML.Lat := 7;
	ML.Lbt := 9;
	ML.Lds := 9;
	ML.Les := 7;
	ML.Lst := 16;

	(*# GC-content of the paired part
	# used to compute kon *)
	ML.fCGbd := 0.5;        (*#this is fake*)
	ML.fCGad := 7.14286E-01;
	ML.fCGbe := 5.71429E-01;
	ML.fCGat := 7.14286E-01;
	ML.fCGbt := 4.44444E-01;
	ML.fCGds := 6.66667E-01;
	ML.fCGes := 5.71429E-01;
	ML.fCGst := 5.0E-01;

	(*# Binding entropies*)
	ML.Sbd := 0.0;        (*#this is fake*)
	ML.Sad := -1.73E-01;
	ML.Sbe := -1.4125E-01;
	ML.Sat := -1.73E-01;
	ML.Sbt := -1.995E-01;
	ML.Sds := -2.2575E-01;
	ML.Ses := -1.4125E-01;
	ML.Sst := -3.9225E-01;

	(*# Binding enthalpies*)
	ML.Hbd := 0.0;        (*#this is fake*)
	ML.Had := -6.4365E+01;
	ML.Hbe := -5.07024E+01;
	ML.Hat := -6.4365E+01;
	ML.Hbt := -7.11634E+01;
	ML.Hds := -8.32036E+01;
	ML.Hes := -5.07024E+01;
	ML.Hst := -1.45603E+02;

	(*# Initial conditions*)
	ML.A0 := 50.E-6;                   (*# M*)
	ML.B0 := 50.E-6;                   (*# M*)
	ML.E0 := 50.E-6;                   (*# M*)
	ML.D0 := 50.E-6;                   (*# M*)
	ML.T0 := 0.5E-6;                   (*# M*)
	ML.S0 := 0.5E-6;                   (*# M*)
	ML.SunY0 := 2.5E-6;                (*# M*)

	(*# Temperature cycling parameters*)
	ML.cycleNum := 1;
	ML.cycleTime := 360.0;            (*# s*)
	ML.heatingRamp := 5.0;             (*# °C/s*)
	ML.hotTime := 1.0;                 (*# s*)
	ML.coolingRamp := 230.0;             (*# °C/s*)
	ML.step := 0.0001;                 (*# s*)
	ML.printPeriod := 10.0;              (*# s*)
	ML.printPeriodHeat := 0.5;         (*# s*)
	ML.tcelsius := 40.0;               (*# reaction temperature, °C*)
	ML.tHeatCelsius := 80.0;           (*# heating temperature, °C*)

	(*# General model parameters*)
	ML.kdeg := 0.002;                 (*# SunY degradation rate at 80°C, 1/s*)
	ML.delayModel := ML.None;             (*# Must be one of: None, Activation, Lag*)
	ML.suppressBDbinding := TRUE;
	ML.degSunY := TRUE;
	ML.Tlag := 2700.0;                 (*# 0.72/3600, SunY fixed activation time,s*)
	ML.kact := 0.000555556;            (*# 1./(45*60), SunY activation rate, 1/s*)
	ML.kEnz := 166.6666667;            (*# 0.3/3600/0.5E-6 <- 1 / M s, = 0.3/h at 0.5 µM SunY*)

	(*# Error tolerances for the adaptive Runge-Kutta method*)
	ML.rtol := 2.0E-3;                 (*# relative error tolerance*)
	ML.atol := 1.0E-8;                 (*# absolute error tolerance, M*)

	(*# General parameters for the kon computations*)
	ML.kUni := 2.1E5;                  (*# 1 / s*)
	ML.kBi := 1.6E7;                   (*# 1 / M s*)
	ML.DeltaGCG := 3.06;               (*# kcal / mol bp*)
	ML.DeltaGAU := 1.32;               (*# kcal / mol bp*)
	ML.DeltaGini := 3.18;              (*# kcal / mol*)
	time := Kernel.Time();
	ML.MainLoop;
	time := Kernel.Time() - time;
	Out.String('Execution time is ');
	Out.Int(time DIV 1000,0);
	Out.Char('.');
	Out.Int(time MOD 1000,0);
	Out.String('s.');
	Out.Ln;
END Run;

END Experiment.

Experiment.Run