MODULE MutualLigationContT;

IMPORT RK, M := Math, Con := Console, Strings;

CONST
    Activation* = 0;
    Lag* = 1;
    None* = 2;

    Rgas = 1.987E-3;    (*  kcal / K mol *)
    kBOverh = 2.8E10;   (* 1 / K s *)
    temp0 = 273.15;     (* K *)

    heating = TRUE;
    cooling = FALSE;

TYPE
    Chemicals = RECORD
        A, B, D, E, S, T,
        BD, AD, BE, AT, BT, DS, ES, ST,
        ABT, DES, ABD, BDE,
        ABDE: REAL
    END;
    DynamicalSystem = POINTER TO DSVars;
    DSVars = RECORD (RK.DynamicalSystemVariables)
        SunY: REAL;
    END;

VAR

    Lbd*, Lad*, Lbe*, Lat*, Lbt*, Lds*, Les*, Lst*: INTEGER;
    fCGbd*, fCGad*, fCGbe*, fCGat*, fCGbt*, fCGds*, fCGes*, fCGst*,
    Sbd*, Sad*, Sbe*, Sat*, Sbt*, Sds*, Ses*, Sst*,
    Hbd*, Had*, Hbe*, Hat*, Hbt*, Hds*, Hes*, Hst*,
    A0*, B0*, E0*, D0*, T0*, S0*, SunY0*,
    cycleTime*, heatingRamp*, hotTime*, coolingRamp*,
    step*, tcelsius*, tHeatCelsius*,
    printPeriod*, printPeriodHeat*,
    kdeg*, Tlag*, kact*, kEnz*, kUni*, kBi*,
    DeltaGCG*, DeltaGAU*, DeltaGini*,
    rtol*, atol*: REAL;
    suppressBDbinding*, degSunY*: BOOLEAN;
    cycleNum*, delayModel*: INTEGER;

    ds: DynamicalSystem;
    kADon, kADoff,
    kATon, kAToff,
    kBDon, kBDoff,  (* <- responsible for the sticky ends *)
    kBEon, kBEoff,
    kBTon, kBToff,
    kSTon, kSToff,
    kDSon, kDSoff,
    kESon, kESoff: REAL;
    t, cycle: INTEGER;
    elapsedPrintPeriods: LONGINT;
    temp, deltaT,
    time,
    SunYtot, SunY,
    timeElapsed, h, hUsed, coolingTime, heatingTime: REAL;

PROCEDURE ComputeChemicals(VAR X: Chemicals; composition: RK.State);
BEGIN
    X.A    := composition[ 0];
    X.B    := composition[ 1];
    X.D    := composition[ 2];
    X.E    := composition[ 3];
    X.S    := composition[ 4];
    X.T    := composition[ 5];
    X.BD   := composition[ 6];
    X.AD   := composition[ 7];
    X.BE   := composition[ 8];
    X.AT   := composition[ 9];
    X.BT   := composition[10];
    X.DS   := composition[11];
    X.ES   := composition[12];
    X.ST   := composition[13];
    X.ABT  := composition[14];
    X.DES  := composition[15];
    X.ABD  := composition[16];
    X.BDE  := composition[17];
    X.ABDE := composition[18];
END ComputeChemicals;

PROCEDURE KD(fCG, temp: REAL): REAL;
BEGIN
    RETURN M.Exp(-2.27*fCG/(0.24 + fCG)/(Rgas*temp))
END KD;

PROCEDURE Kd(S, H, temp: REAL): REAL;
BEGIN
    RETURN M.Exp((H/temp - S)/Rgas)
END Kd;

PROCEDURE kon(L: INTEGER; fCG, temp: REAL): REAL;
BEGIN
    RETURN L*kBi*2*kUni/(2*kUni + kBi*KD(fCG,temp))
END kon;

PROCEDURE ComputeK(temp: REAL);
BEGIN
    kADon  := kon(Lad,fCGad,temp); (**)
    kADoff := kADon*Kd(Sad,Had,temp);

    kATon  := kon(Lat,fCGat,temp); (**)
    kAToff := kATon*Kd(Sat,Hat,temp);

    IF suppressBDbinding THEN
    kBDon  := 0.0; (**)
    kBDoff := 0.0;
    ELSE
    kBDon  := kon(Lbd,fCGbd,temp); (**)
    kBDoff := kBDon*Kd(Sbd,Hbd,temp);
    END;

    kBEon  := kon(Lbe,fCGbe,temp); (**)
    kBEoff := kBEon*Kd(Sbe,Hbe,temp);

    kBTon  := kon(Lbt,fCGbt,temp); (**)
    kBToff := kBTon*Kd(Sbt,Hbt,temp);

    kSTon  := kon(Lst,fCGst,temp); (**)
    kSToff := kSTon*Kd(Sst,Hst,temp);

    kDSon  := kon(Lds,fCGds,temp); (**)
    kDSoff := kDSon*Kd(Sds,Hds,temp);

    kESon  := kon(Les,fCGes,temp); (**)
    kESoff := kESon*Kd(Ses,Hes,temp);
END ComputeK;

PROCEDURE PrintHeadder;
BEGIN
    Con.WriteStr("     cycle ");
    Con.WriteStr("           t ");
    Con.WriteStr("        [A] ");
    Con.WriteStr("        [B] ");
    Con.WriteStr("        [D] ");
    Con.WriteStr("        [E] ");
    Con.WriteStr("        [S] ");
    Con.WriteStr("        [T] ");
    Con.WriteStr("       [BD] ");
    Con.WriteStr("       [AD] ");
    Con.WriteStr("       [BE] ");
    Con.WriteStr("       [AT] ");
    Con.WriteStr("       [BT] ");
    Con.WriteStr("       [DS] ");
    Con.WriteStr("       [ES] ");
    Con.WriteStr("       [ST] ");
    Con.WriteStr("      [ABT] ");
    Con.WriteStr("      [DES] ");
    Con.WriteStr("      [ABD] ");
    Con.WriteStr("      [BDE] ");
    Con.WriteStr("     [ABDE] ");
    Con.WriteStr("     [Atot] ");
    Con.WriteStr("     [Btot] ");
    Con.WriteStr("     [Dtot] ");
    Con.WriteStr("     [Etot] ");
    Con.WriteStr("     [Stot] ");
    Con.WriteStr("     [Ttot] ");
    Con.WriteStr("  [SunYtot] ");
    Con.WriteStr("     [SunY] ");
    Con.WriteStr("    [Sprod] ");
    Con.WriteStr("    [Tprod] ");
    Con.WriteLn;
END PrintHeadder;

PROCEDURE (ds: DynamicalSystem) ComputeRHS(VAR result: RK.State);
VAR X: Chemicals;
BEGIN
    ComputeChemicals(X, ds.state);
    result[ 0] := (* A *)
        - kADon*X.A*X.D   + kADoff*X.AD   - kATon*X.A*X.T  + kAToff*X.AT
        - kADon*X.A*X.BD  + kADoff*X.ABD  - kATon*X.A*X.BT + kAToff*X.ABT
        - kADon*X.A*X.BDE + kADoff*X.ABDE;
    result[ 1] := (* B *)
        - kBDon*X.B*X.D  + kBDoff*X.BD  - kBEon*X.B*X.E  + kBEoff*X.BE
        - kBTon*X.B*X.T  + kBToff*X.BT  - kBTon*X.B*X.AT + kBToff*X.ABT
        - kBDon*X.B*X.AD + kBDoff*X.ABD;
    result[ 2] := (* D *)
        - kADon*X.A*X.D  + kADoff*X.AD  - kBDon*X.B*X.D  + kBDoff*X.BD
        - kDSon*X.D*X.S  + kDSoff*X.DS  - kDSon*X.D*X.ES + kDSoff*X.DES
        - kBDon*X.D*X.BE + kBDoff*X.BDE;
    result[ 3] := (* E *)
        - kBEon*X.E*X.B   + kBEoff*X.BE   - kESon*X.E*X.S  + kESoff*X.ES
        - kBEon*X.E*X.BD  + kBEoff*X.BDE  - kESon*X.E*X.DS + kESoff*X.DES
        - kBEon*X.E*X.ABD + kBEoff*X.ABDE;
    result[ 4] := (* S *)
        - kSTon*X.S*X.T + kSToff*X.ST - kDSon*X.S*X.D + kDSoff*X.DS
        - kESon*X.S*X.E + kESoff*X.ES;
    result[ 5] := (* T *)
        - kSTon*X.T*X.S + kSToff*X.ST - kBTon*X.T*X.B + kBToff*X.BT
        - kATon*X.T*X.A + kAToff*X.AT;
    result[ 6] := (* BD *)
        - kADon*X.A*X.BD + kADoff*X.ABD + kBDon*X.B*X.D - kBDoff*X.BD
        - kBEon*X.E*X.BD + kBEoff*X.BDE;
    result[ 7] := (* AD *)
          kADon*X.A*X.D - kADoff*X.AD - kBDon*X.B*X.AD + kBDoff*X.ABD
        - kBDon*X.AD*X.BE + kBDoff*X.ABDE;
    result[ 8] := (* BE *)
          kBEon*X.B*X.E   - kBEoff*X.BE   - kBDon*X.D*X.BE + kBDoff*X.BDE
        - kBDon*X.AD*X.BE + kBDoff*X.ABDE;
    result[ 9] := (* AT *)
          kATon*X.A*X.T - kAToff*X.AT - kBTon*X.B*X.AT + kBToff*X.ABT;
    result[10] := (* BT *)
        - kATon*X.A*X.BT + kAToff*X.ABT + kBTon*X.B*X.T - kBToff*X.BT
        + kEnz*ds.SunY*X.BDE;
    result[11] := (* DS *)
          kDSon*X.D*X.S - kDSoff*X.DS - kESon*X.E*X.DS + kESoff*X.DES
        + kEnz*ds.SunY*X.ABD;
    result[12] := (* ES *)
        - kDSon*X.D*X.ES + kDSoff*X.DES + kESon*X.E*X.S - kESoff*X.ES;
    result[13] := (* ST *)
          kSTon*X.S*X.T - kSToff*X.ST
        + kEnz*ds.SunY*(X.ABT + X.DES);
    result[14] := (* ABT *)
          kATon*X.A*X.BT - kAToff*X.ABT + kBTon*X.B*X.AT - kBToff*X.ABT
        + kEnz*ds.SunY*(X.ABDE/2.0 - X.ABT);
    result[15] := (* DES *)
          kDSon*X.D*X.ES - kDSoff*X.DES + kESon*X.E*X.DS - kESoff*X.DES
        + kEnz*ds.SunY*(X.ABDE/2.0 - X.DES);
    result[16] := (* ABD *)
          kADon*X.A*X.BD  - kADoff*X.ABD + kBDon*X.B*X.AD - kBDoff*X.ABD
        - kBEon*X.E*X.ABD + kBEoff*X.ABDE
        - kEnz*ds.SunY*X.ABD;
    result[17] := (* BDE *)
        - kADon*X.A*X.BDE + kADoff*X.ABDE + kBDon*X.D*X.BE - kBDoff*X.BDE
        + kBEon*X.E*X.BD  - kBEoff*X.BDE
        - kEnz*ds.SunY*X.BDE;
    result[18] := (* ABDE *)
          kADon*X.A*X.BDE - kADoff*X.ABDE + kBEon*X.E*X.ABD - kBEoff*X.ABDE
        + kBDon*X.AD*X.BE - kBDoff*X.ABDE
        - kEnz*ds.SunY*X.ABDE;
END ComputeRHS;

PROCEDURE TestZeroStep;
BEGIN
    IF hUsed <= 1/MAX(REAL) THEN
        Con.WriteStr('# The simulation terminated with the estimated step = 0');
        Con.WriteLn;
        HALT(0);
    END
END TestZeroStep;

PROCEDURE Print(composition: RK.State; cycle: INTEGER; time: REAL);
VAR
    X: Chemicals;
    Atot, Btot, Dtot, Etot, Stot, Ttot: REAL;
    i: INTEGER;
    s: ARRAY 512 OF CHAR;

    PROCEDURE Format(number: REAL; precision: INTEGER);
    BEGIN
        Strings.RealToStringForm(number,precision,precision+6,2,' ',s);
        Con.WriteStr(s);
    END Format;

BEGIN
    ComputeChemicals(X, composition);
    Stot := X.S + X.DS + X.ES + X.DES + X.ST;
    Ttot := X.T + X.AT + X.BT + X.ABT + X.ST;
    Atot := A0 + S0 - Stot;
    Btot := B0 + S0 - Stot;
    Dtot := D0 + T0 - Ttot;
    Etot := E0 + T0 - Ttot;
    Strings.IntToStringForm(cycle+1,Strings.decimal,10,' ',Strings.hideBase,s);
    Con.WriteStr(s);
    Format(time,7);
    FOR i := 0 TO 18 DO
        Format(composition[i],6);
    END;
    Format(Atot,6);
    Format(Btot,6);
    Format(Dtot,6);
    Format(Etot,6);
    Format(Stot,6);
    Format(Ttot,6);
    Format(SunYtot,6);
    Format(SunY,6);
    Format(Stot - S0,6);
    Format(Ttot - T0,6);
    Con.WriteLn;
END Print;

PROCEDURE PerformReaction;
BEGIN
    time := 0.0;
    temp := temp0 + tcelsius;
    ComputeK(temp);
    elapsedPrintPeriods := 0;
    WHILE time < cycleTime DO
        IF delayModel = Lag THEN IF time < Tlag THEN SunY := 0. ELSE SunY := SunYtot END;
        ELSIF delayModel = Activation THEN SunY := SunYtot*(1 - M.Exp(-kact*time));
        ELSE SunY := SunYtot;
        END;
        ds.SunY := SunY;
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriod) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriod);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END PerformReaction;

PROCEDURE KeepHot;
BEGIN
    time := 0.0;
    temp := temp0 + tHeatCelsius;
    ComputeK(temp);
    ds.SunY := 0.0;
    elapsedPrintPeriods := 0;
    WHILE time < hotTime DO
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriodHeat) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriodHeat);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END KeepHot;

PROCEDURE ChangeTemperature(heatingMode: BOOLEAN);
VAR initT, deltaT, targetTime: REAL;
BEGIN
    time := 0.0;
    ds.SunY := 0.0;
    elapsedPrintPeriods := 0;
    IF heatingMode = heating THEN
        initT := temp0 + tcelsius;
        deltaT := tHeatCelsius - tcelsius;
        targetTime := heatingTime;
    ELSE
        initT := temp0 + tHeatCelsius;
        deltaT := tcelsius - tHeatCelsius;
        targetTime := coolingTime;
    END;
    WHILE time < targetTime DO
        temp := initT + deltaT*time/targetTime;
        ComputeK(temp);
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriodHeat) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriodHeat);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END ChangeTemperature;

PROCEDURE MainLoop*;
BEGIN
    NEW(ds); NEW(ds.state,19);
    ds.state[0] := A0; ds.state[1] := B0;
    ds.state[2] := D0; ds.state[3] := E0;
    ds.state[4] := S0; ds.state[5] := T0;
    FOR t := 6 TO 18 DO
        ds.state[t] := 0.0;
    END;
    SunYtot := SunY0; SunY := SunYtot;
    PrintHeadder;
    Print(ds.state,0,0);
    timeElapsed := 0.0;
    h := step;
    heatingTime := (tHeatCelsius - tcelsius)/heatingRamp;
    coolingTime := (tHeatCelsius - tcelsius)/coolingRamp;
    FOR cycle := 0 TO cycleNum - 1 DO
        ChangeTemperature(heating);
        KeepHot;
        ChangeTemperature(cooling);
        IF degSunY THEN SunYtot := SunYtot*M.Exp(-kdeg*(hotTime+heatingTime+coolingTime)) END;
        PerformReaction;
    END;
END MainLoop;

END MutualLigationContT.
