#import Pkg
#Pkg.add("OrdinaryDiffEq")
using OrdinaryDiffEq

const Rgas = 1.987e-3 # ccal / K mol
const temp0 = 273.15 # K

const Lbd=1.0        #this is fake
const Lad=7.0
const Lbe=7.0
const Lat=7.0
const Lbt=9.0
const Lds=9.0
const Les=7.0
const Lst=16.0

# GC-content of the paired part
# used to compute kon
const fCGbd=0.5        #this is fake
const fCGad=7.14286e-01
const fCGbe=5.71429e-01
const fCGat=7.14286e-01
const fCGbt=4.44444e-01
const fCGds=6.66667e-01
const fCGes=5.71429e-01
const fCGst=5.0e-01

# Binding entropies
const Sbd=0.0        #this is fake
const Sad=-1.73e-01
const Sbe=-1.4125e-01
const Sat=-1.73e-01
const Sbt=-1.995e-01
const Sds=-2.2575e-01
const Ses=-1.4125e-01
const Sst=-3.9225e-01

# Binding enthalpies
const Hbd=0.0        #this is fake
const Had=-6.4365e+01
const Hbe=-5.07024e+01
const Hat=-6.4365e+01
const Hbt=-7.11634e+01
const Hds=-8.32036e+01
const Hes=-5.07024e+01
const Hst=-1.45603e+02

# Initial conditions
const A0=50.e-6                   # M
const B0=50.e-6                   # M
const E0=50.e-6                   # M
const D0=50.e-6                   # M
const T0=0.5e-6                   # M
const S0=0.5e-6                   # M
const SunY0=2.5e-6                # M

# Temperature cycling parameters
const cycleNum=1.0
const cycleTime=360.0            # s
const heatingRamp=5.0             # °C/s
const hotTime=1.0                 # s
const coolingRamp=230.0             # °C/s
const step=0.0001                 # s
const printPeriod=10.0              # s
const printPeriodHeat=0.5         # s
const tcelsius=40.0               # reaction temperature, °C
const tHeatCelsius=80.0           # heating temperature, °C

# General model parameters
const kdeg=0.002                 # SunY degradation rate at 80°C, 1/s
const kact=0.000555556            # 1./(45*60), SunY activation rate, 1/s
const kEnz=166.6666667            # 0.3/3600/0.5e-6 <- 1 / M s, = 0.3/h at 0.5 µM SunY

# Error tolerances for the adaptive Runge-Kutta method
const rtol=2.0e-3                 # relative error tolerance
const atol=1.0e-8                 # absolute error tolerance, M

# General parameters for the kon computations
const kUni=2.1E5                  # 1 / s
const kBi=1.6E7                   # 1 / M s
const DeltaGCG=3.06               # kcal / mol bp
const DeltaGAU=1.32               # kcal / mol bp
const DeltaGini=3.18              # kcal / mol

KD(fCG,temp) = exp(-2.27*fCG/(0.24 + fCG)/(Rgas*temp))

Kd(S,H,temp) = exp((H/temp - S)/Rgas)

kon(L,fCG,temp) = L*kBi*2*kUni/(2*kUni + kBi*KD(fCG,temp))

function RHS(result, y, p, t)
	tempIni = p[1]
	dtemp = p[2]
	temp = temp0 + tempIni + dtemp*t

	kADon  = kon(Lad,fCGad,temp) #(**)
	kADoff = kADon*Kd(Sad,Had,temp)

	kATon  = kon(Lat,fCGat,temp) #(**)
	kAToff = kATon*Kd(Sat,Hat,temp)

	kBDon  = 0.0 #(**)
	kBDoff = 0.0

	kBEon  = kon(Lbe,fCGbe,temp) #(**)
	kBEoff = kBEon*Kd(Sbe,Hbe,temp)

	kBTon  = kon(Lbt,fCGbt,temp) #(**)
	kBToff = kBTon*Kd(Sbt,Hbt,temp)

	kSTon  = kon(Lst,fCGst,temp) #(**)
	kSToff = kSTon*Kd(Sst,Hst,temp)

	kDSon  = kon(Lds,fCGds,temp) #(**)
	kDSoff = kDSon*Kd(Sds,Hds,temp)

	kESon  = kon(Les,fCGes,temp); #(**)
	kESoff = kESon*Kd(Ses,Hes,temp)

	result[19] = 	- kADon*y[19]*y[2]   + kADoff*y[7]   - kATon*y[19]*y[5]  + kAToff*y[9] 	- kADon*y[19]*y[6]  + kADoff*y[16]  - kATon*y[19]*y[10] + kAToff*y[14] 	- kADon*y[19]*y[17] + kADoff*y[18]
	result[ 1] = 	- kBDon*y[1]*y[2]  + kBDoff*y[6]  - kBEon*y[1]*y[3]  + kBEoff*y[8] 	- kBTon*y[1]*y[5]  + kBToff*y[10]  - kBTon*y[1]*y[9] + kBToff*y[14] 	- kBDon*y[1]*y[7] + kBDoff*y[16]
	result[ 2] = 	- kADon*y[19]*y[2]  + kADoff*y[7]  - kBDon*y[1]*y[2]  + kBDoff*y[6] 	- kDSon*y[2]*y[4]  + kDSoff*y[11]  - kDSon*y[2]*y[12] + kDSoff*y[15] 	- kBDon*y[2]*y[8] + kBDoff*y[17]
	result[ 3] = 	- kBEon*y[3]*y[1]   + kBEoff*y[8]   - kESon*y[3]*y[4]  + kESoff*y[12] 	- kBEon*y[3]*y[6]  + kBEoff*y[17]  - kESon*y[3]*y[11] + kESoff*y[15] 	- kBEon*y[3]*y[16] + kBEoff*y[18]
	result[ 4] = 	- kSTon*y[4]*y[5] + kSToff*y[13] - kDSon*y[4]*y[2] + kDSoff*y[11] 	- kESon*y[4]*y[3] + kESoff*y[12]
	result[ 5] = 	- kSTon*y[5]*y[4] + kSToff*y[13] - kBTon*y[5]*y[1] + kBToff*y[10] 	- kATon*y[5]*y[19] + kAToff*y[9]
	result[ 6] = 	- kADon*y[19]*y[6] + kADoff*y[16] + kBDon*y[1]*y[2] - kBDoff*y[6] 	- kBEon*y[3]*y[6] + kBEoff*y[17]
	result[ 7] = 	  kADon*y[19]*y[2] - kADoff*y[7] - kBDon*y[1]*y[7] + kBDoff*y[16] 	- kBDon*y[7]*y[8] + kBDoff*y[18]
	result[ 8] = 	  kBEon*y[1]*y[3]   - kBEoff*y[8]   - kBDon*y[2]*y[8] + kBDoff*y[17] 	- kBDon*y[7]*y[8] + kBDoff*y[18]
	result[ 9] = 	  kATon*y[19]*y[5] - kAToff*y[9] - kBTon*y[1]*y[9] + kBToff*y[14]
	result[10] = 	- kATon*y[19]*y[10] + kAToff*y[14] + kBTon*y[1]*y[5] - kBToff*y[10] 	+ kEnz*SunY*y[17]
	result[11] = 	  kDSon*y[2]*y[4] - kDSoff*y[11] - kESon*y[3]*y[11] + kESoff*y[15] 	+ kEnz*SunY*y[16]
	result[12] = 	- kDSon*y[2]*y[12] + kDSoff*y[15] + kESon*y[3]*y[4] - kESoff*y[12]
	result[13] = 	  kSTon*y[4]*y[5] - kSToff*y[13] 	+ kEnz*SunY*(y[14] + y[15])
	result[14] = 	  kATon*y[19]*y[10] - kAToff*y[14] + kBTon*y[1]*y[9] - kBToff*y[14] 	+ kEnz*SunY*(y[18]/2.0 - y[14])
	result[15] = 	  kDSon*y[2]*y[12] - kDSoff*y[15] + kESon*y[3]*y[11] - kESoff*y[15] 	+ kEnz*SunY*(y[18]/2.0 - y[15])
	result[16] = 	  kADon*y[19]*y[6]  - kADoff*y[16] + kBDon*y[1]*y[7] - kBDoff*y[16] 	- kBEon*y[3]*y[16] + kBEoff*y[18] 	- kEnz*SunY*y[16]
	result[17] = 	- kADon*y[19]*y[17] + kADoff*y[18] + kBDon*y[2]*y[8] - kBDoff*y[17] 	+ kBEon*y[3]*y[6]  - kBEoff*y[17] 	- kEnz*SunY*y[17]
	result[18] = 	  kADon*y[19]*y[17] - kADoff*y[18] + kBEon*y[3]*y[16] - kBEoff*y[18] 	+ kBDon*y[7]*y[8] - kBDoff*y[18] 	- kEnz*SunY*y[18]
end

SunY = 0.0

u01 = [B0, D0, E0, S0, T0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, A0]
p1 = [tcelsius,heatingRamp]
tspan1 = (tHeatCelsius - tcelsius)/heatingRamp
prob1 = ODEProblem(RHS,u01,tspan1,p1)
sol1 = solve(prob1,Tsit5(),reltol=rtol,abstol=atol,saveat=printPeriodHeat)
print(sol1.t)
print(sol1.u)

u02 = last(sol1.u)
p2 = [tHeatCelsius,0.0]
tspan2 = hotTime
prob2 = ODEProblem(RHS,u02,tspan2,p2)
sol2 = solve(prob2,Tsit5(),reltol=rtol,abstol=atol,saveat=printPeriodHeat)
print(sol2.t)
print(sol2.u)

u03 = last(sol2.u)
p3 = [tHeatCelsius,-coolingRamp]
tspan3 = (tHeatCelsius - tcelsius)/coolingRamp
prob3 = ODEProblem(RHS,u03,tspan3,p3)
sol3 = solve(prob3,Tsit5(),reltol=rtol,abstol=atol,saveat=printPeriodHeat)
print(sol3.t)
print(sol3.u)

SunY = SunY0*exp(-kdeg*(tspan1+tspan2+tspan3))

u04 = last(sol3.u)
p4 = [tcelsius,0.0]
tspan4 = cycleTime
prob4 = ODEProblem(RHS,u04,tspan4,p4)
sol4 = solve(prob4,Tsit5(),reltol=rtol,abstol=atol,saveat=printPeriod)
print(sol4.t)
print(sol4.u)
