MODULE MutualLigationContT;

IMPORT RK, M := MathL, Out, Args := CmdArgs, Strings := bbStrings, SYSTEM, Platform;

CONST
    Activation = 0;
    Lag = 1;
    None = 2;

    Rgas = 1.987D-3;    (*  kcal / K mol *)
    kBOverh = 2.8D10;   (* 1 / K s *)
    temp0 = 273.15;     (* K *)

    heating = TRUE;
    cooling = FALSE;

TYPE
    Chemicals = RECORD
        A, B, D, E, S, T,
        BD, AD, BE, AT, BT, DS, ES, ST,
        ABT, DES, ABD, BDE,
        ABDE: LONGREAL
    END;
    DynamicalSystem = POINTER TO DSVars;
    DSVars = RECORD (RK.DynamicalSystemVariables)
        SunY: LONGREAL;
    END;

VAR

    Lbd, Lad, Lbe, Lat, Lbt, Lds, Les, Lst: LONGINT;
    fCGbd, fCGad, fCGbe, fCGat, fCGbt, fCGds, fCGes, fCGst,
    Sbd, Sad, Sbe, Sat, Sbt, Sds, Ses, Sst,
    Hbd, Had, Hbe, Hat, Hbt, Hds, Hes, Hst,
    A0, B0, E0, D0, T0, S0, SunY0,
    cycleTime, heatingRamp, hotTime, coolingRamp,
    step, tcelsius, tHeatCelsius,
    printPeriod, printPeriodHeat,
    kdeg, Tlag, kact, kEnz, kUni, kBi,
    DeltaGCG, DeltaGAU, DeltaGini,
    rtol, atol: LONGREAL;
    suppressBDbinding, degSunY: BOOLEAN;
    cycleNum, delayModel: LONGINT;

    ds: DynamicalSystem;
    kADon, kADoff,
    kATon, kAToff,
    kBDon, kBDoff,  (* <- responsible for the sticky ends *)
    kBEon, kBEoff,
    kBTon, kBToff,
    kSTon, kSToff,
    kDSon, kDSoff,
    kESon, kESoff: LONGREAL;
    t, cycle, elapsedPrintPeriods: LONGINT;
    temp, deltaT,
    time,
    SunYtot, SunY,
    timeElapsed, h, hUsed, coolingTime, heatingTime: LONGREAL;

PROCEDURE GetParameters():BOOLEAN;
VAR
    res, i, j, m: LONGINT; result: BOOLEAN;
    s: ARRAY 64 OF CHAR;
    s16: ARRAY 64 OF SYSTEM.CHAR16;
    a: ARRAY 51 OF LONGREAL;
    b: ARRAY 9 OF LONGINT;
    r: LONGREAL;
BEGIN
    result := Args.Count = 63;
    FOR i := 0 TO 50 DO
        IF result THEN
            Args.Get(i+1,s); Strings.Utf8ToString(s,s16,res);
            Strings.StringToReal(s16,r,res);
            IF res = 0 THEN a[i] := r;
            ELSE result := FALSE END;
        END
    END;
    FOR i := 0 TO 8 DO
        IF result THEN
            Args.Get(i+52,s); Strings.Utf8ToString(s,s16,res);
            Strings.StringToInt(s16,m,res);
            IF res = 0 THEN b[i] := m;
            ELSE result := FALSE END;
        END;
    END;
    IF result THEN
        fCGbd:=a[0]; fCGad:=a[1]; fCGbe:=a[2]; fCGat:=a[3];
        fCGbt:=a[4]; fCGds:=a[5]; fCGes:=a[6]; fCGst:=a[7];
        Sbd:=a[8]; Sad:=a[9]; Sbe:=a[10]; Sat:=a[11];
        Sbt:=a[12]; Sds:=a[13]; Ses:=a[14]; Sst:=a[15];
        Hbd:=a[16]; Had:=a[17]; Hbe:=a[18]; Hat:=a[19];
        Hbt:=a[20]; Hds:=a[21]; Hes:=a[22]; Hst:=a[23];
        A0:=a[24]; B0:=a[25]; E0:=a[26]; D0:=a[27];
        T0:=a[28]; S0:=a[29]; SunY0:=a[30];
        cycleTime:=a[31]; heatingRamp:=a[32];
        hotTime:=a[33]; coolingRamp:=a[34];
        step:=a[35]; tcelsius:=a[36]; tHeatCelsius:=a[37];
        printPeriod:=a[38]; printPeriodHeat:=a[39];
        kdeg:=a[40]; Tlag:=a[41]; kact:=a[42]; kEnz:=a[43];
        kUni:=a[44]; kBi:=a[45];
        DeltaGCG:=a[46]; DeltaGAU:=a[47]; DeltaGini:=a[48];
        rtol:=a[49]; atol:=a[50];
        Lbd:=b[0]; Lad:=b[1]; Lbe:=b[2]; Lat:=b[3];
        Lbt:=b[4]; Lds:=b[5]; Les:=b[6]; Lst:=b[7];
        cycleNum:=b[8]
    END;
    IF result THEN Args.Get(61,s);
        IF s = 'TRUE' THEN suppressBDbinding := TRUE;
        ELSIF s = 'FALSE' THEN suppressBDbinding := FALSE;
        ELSE result := FALSE END
    END;
    IF result THEN Args.Get(62,s);
        IF s = 'TRUE' THEN degSunY := TRUE;
        ELSIF s = 'FALSE' THEN degSunY := FALSE;
        ELSE result := FALSE END
    END;
    IF result THEN
        Args.Get(63,s);
        IF s = 'None' THEN delayModel := None;
        ELSIF s = 'Activation' THEN delayModel := Activation;
        ELSIF s = 'Lag' THEN delayModel := Lag;
        ELSE result := FALSE
        END
    END;
    RETURN result
END GetParameters;

PROCEDURE ComputeChemicals(VAR X: Chemicals; composition: RK.State);
BEGIN
    X.A    := composition[ 0];
    X.B    := composition[ 1];
    X.D    := composition[ 2];
    X.E    := composition[ 3];
    X.S    := composition[ 4];
    X.T    := composition[ 5];
    X.BD   := composition[ 6];
    X.AD   := composition[ 7];
    X.BE   := composition[ 8];
    X.AT   := composition[ 9];
    X.BT   := composition[10];
    X.DS   := composition[11];
    X.ES   := composition[12];
    X.ST   := composition[13];
    X.ABT  := composition[14];
    X.DES  := composition[15];
    X.ABD  := composition[16];
    X.BDE  := composition[17];
    X.ABDE := composition[18];
END ComputeChemicals;

PROCEDURE KD(fCG, temp: LONGREAL): LONGREAL;
BEGIN
    RETURN M.exp(-2.27*fCG/(0.24 + fCG)/(Rgas*temp))
END KD;

PROCEDURE Kd(S, H, temp: LONGREAL): LONGREAL;
BEGIN
    RETURN M.exp((H/temp - S)/Rgas)
END Kd;

PROCEDURE kon(L: LONGINT; fCG, temp: LONGREAL): LONGREAL;
BEGIN
    RETURN L*kBi*2*kUni/(2*kUni + kBi*KD(fCG,temp))
END kon;

PROCEDURE ComputeK(temp: LONGREAL);
BEGIN
    kADon  := kon(Lad,fCGad,temp); (**)
    kADoff := kADon*Kd(Sad,Had,temp);

    kATon  := kon(Lat,fCGat,temp); (**)
    kAToff := kATon*Kd(Sat,Hat,temp);

    IF suppressBDbinding THEN
    kBDon  := 0.0; (**)
    kBDoff := 0.0;
    ELSE
    kBDon  := kon(Lbd,fCGbd,temp); (**)
    kBDoff := kBDon*Kd(Sbd,Hbd,temp);
    END;

    kBEon  := kon(Lbe,fCGbe,temp); (**)
    kBEoff := kBEon*Kd(Sbe,Hbe,temp);

    kBTon  := kon(Lbt,fCGbt,temp); (**)
    kBToff := kBTon*Kd(Sbt,Hbt,temp);

    kSTon  := kon(Lst,fCGst,temp); (**)
    kSToff := kSTon*Kd(Sst,Hst,temp);

    kDSon  := kon(Lds,fCGds,temp); (**)
    kDSoff := kDSon*Kd(Sds,Hds,temp);

    kESon  := kon(Les,fCGes,temp); (**)
    kESoff := kESon*Kd(Ses,Hes,temp);
END ComputeK;

(*
PROCEDURE PrintK;
BEGIN
    Out.String('ADon/off'); Out.Ln;
    Out.Real(SHORT(kADon),0); Out.Ln;
    Out.Real(SHORT(kADoff),0); Out.Ln; Out.Ln;
    Out.String('ATon/off'); Out.Ln;
    Out.Real(SHORT(kATon),0); Out.Ln;
    Out.Real(SHORT(kAToff),0); Out.Ln; Out.Ln;
    Out.String('BDon/off'); Out.Ln;
    Out.Real(SHORT(kBDon),0); Out.Ln;
    Out.Real(SHORT(kBDoff),0); Out.Ln; Out.Ln;
    Out.String('BEon/off'); Out.Ln;
    Out.Real(SHORT(kBEon),0); Out.Ln;
    Out.Real(SHORT(kBEoff),0); Out.Ln; Out.Ln;
    Out.String('BTon/off'); Out.Ln;
    Out.Real(SHORT(kBTon),0); Out.Ln;
    Out.Real(SHORT(kBToff),0); Out.Ln; Out.Ln;
    Out.String('STon/off'); Out.Ln;
    Out.Real(SHORT(kSTon),0); Out.Ln;
    Out.Real(SHORT(kSToff),0); Out.Ln; Out.Ln;
    Out.String('DSon/off'); Out.Ln;
    Out.Real(SHORT(kDSon),0); Out.Ln;
    Out.Real(SHORT(kDSoff),0); Out.Ln; Out.Ln;
    Out.String('ESon/off'); Out.Ln;
    Out.Real(SHORT(kESon),0); Out.Ln;
    Out.Real(SHORT(kESoff),0); Out.Ln; Out.Ln;
END PrintK;
*)

PROCEDURE PrintHeadder;
BEGIN
    Out.String("     cycle ");
    Out.String("           t ");
    Out.String("        [A] ");
    Out.String("        [B] ");
    Out.String("        [D] ");
    Out.String("        [E] ");
    Out.String("        [S] ");
    Out.String("        [T] ");
    Out.String("       [BD] ");
    Out.String("       [AD] ");
    Out.String("       [BE] ");
    Out.String("       [AT] ");
    Out.String("       [BT] ");
    Out.String("       [DS] ");
    Out.String("       [ES] ");
    Out.String("       [ST] ");
    Out.String("      [ABT] ");
    Out.String("      [DES] ");
    Out.String("      [ABD] ");
    Out.String("      [BDE] ");
    Out.String("     [ABDE] ");
    Out.String("     [Atot] ");
    Out.String("     [Btot] ");
    Out.String("     [Dtot] ");
    Out.String("     [Etot] ");
    Out.String("     [Stot] ");
    Out.String("     [Ttot] ");
    Out.String("  [SunYtot] ");
    Out.String("     [SunY] ");
    Out.String("    [Sprod] ");
    Out.String("    [Tprod] ");
    Out.Ln
END PrintHeadder;

PROCEDURE (ds: DynamicalSystem) ComputeRHS(VAR result: RK.State);
VAR X: Chemicals;
BEGIN
    ComputeChemicals(X, ds.state);
    result[ 0] := (* A *)
        - kADon*X.A*X.D   + kADoff*X.AD   - kATon*X.A*X.T  + kAToff*X.AT
        - kADon*X.A*X.BD  + kADoff*X.ABD  - kATon*X.A*X.BT + kAToff*X.ABT
        - kADon*X.A*X.BDE + kADoff*X.ABDE;
    result[ 1] := (* B *)
        - kBDon*X.B*X.D  + kBDoff*X.BD  - kBEon*X.B*X.E  + kBEoff*X.BE
        - kBTon*X.B*X.T  + kBToff*X.BT  - kBTon*X.B*X.AT + kBToff*X.ABT
        - kBDon*X.B*X.AD + kBDoff*X.ABD;
    result[ 2] := (* D *)
        - kADon*X.A*X.D  + kADoff*X.AD  - kBDon*X.B*X.D  + kBDoff*X.BD
        - kDSon*X.D*X.S  + kDSoff*X.DS  - kDSon*X.D*X.ES + kDSoff*X.DES
        - kBDon*X.D*X.BE + kBDoff*X.BDE;
    result[ 3] := (* E *)
        - kBEon*X.E*X.B   + kBEoff*X.BE   - kESon*X.E*X.S  + kESoff*X.ES
        - kBEon*X.E*X.BD  + kBEoff*X.BDE  - kESon*X.E*X.DS + kESoff*X.DES
        - kBEon*X.E*X.ABD + kBEoff*X.ABDE;
    result[ 4] := (* S *)
        - kSTon*X.S*X.T + kSToff*X.ST - kDSon*X.S*X.D + kDSoff*X.DS
        - kESon*X.S*X.E + kESoff*X.ES;
    result[ 5] := (* T *)
        - kSTon*X.T*X.S + kSToff*X.ST - kBTon*X.T*X.B + kBToff*X.BT
        - kATon*X.T*X.A + kAToff*X.AT;
    result[ 6] := (* BD *)
        - kADon*X.A*X.BD + kADoff*X.ABD + kBDon*X.B*X.D - kBDoff*X.BD
        - kBEon*X.E*X.BD + kBEoff*X.BDE;
    result[ 7] := (* AD *)
          kADon*X.A*X.D - kADoff*X.AD - kBDon*X.B*X.AD + kBDoff*X.ABD
        - kBDon*X.AD*X.BE + kBDoff*X.ABDE;
    result[ 8] := (* BE *)
          kBEon*X.B*X.E   - kBEoff*X.BE   - kBDon*X.D*X.BE + kBDoff*X.BDE
        - kBDon*X.AD*X.BE + kBDoff*X.ABDE;
    result[ 9] := (* AT *)
          kATon*X.A*X.T - kAToff*X.AT - kBTon*X.B*X.AT + kBToff*X.ABT;
    result[10] := (* BT *)
        - kATon*X.A*X.BT + kAToff*X.ABT + kBTon*X.B*X.T - kBToff*X.BT
        + kEnz*ds.SunY*X.BDE;
    result[11] := (* DS *)
          kDSon*X.D*X.S - kDSoff*X.DS - kESon*X.E*X.DS + kESoff*X.DES
        + kEnz*ds.SunY*X.ABD;
    result[12] := (* ES *)
        - kDSon*X.D*X.ES + kDSoff*X.DES + kESon*X.E*X.S - kESoff*X.ES;
    result[13] := (* ST *)
          kSTon*X.S*X.T - kSToff*X.ST
        + kEnz*ds.SunY*(X.ABT + X.DES);
    result[14] := (* ABT *)
          kATon*X.A*X.BT - kAToff*X.ABT + kBTon*X.B*X.AT - kBToff*X.ABT
        + kEnz*ds.SunY*(X.ABDE/2.0 - X.ABT);
    result[15] := (* DES *)
          kDSon*X.D*X.ES - kDSoff*X.DES + kESon*X.E*X.DS - kESoff*X.DES
        + kEnz*ds.SunY*(X.ABDE/2.0 - X.DES);
    result[16] := (* ABD *)
          kADon*X.A*X.BD  - kADoff*X.ABD + kBDon*X.B*X.AD - kBDoff*X.ABD
        - kBEon*X.E*X.ABD + kBEoff*X.ABDE
        - kEnz*ds.SunY*X.ABD;
    result[17] := (* BDE *)
        - kADon*X.A*X.BDE + kADoff*X.ABDE + kBDon*X.D*X.BE - kBDoff*X.BDE
        + kBEon*X.E*X.BD  - kBEoff*X.BDE
        - kEnz*ds.SunY*X.BDE;
    result[18] := (* ABDE *)
          kADon*X.A*X.BDE - kADoff*X.ABDE + kBEon*X.E*X.ABD - kBEoff*X.ABDE
        + kBDon*X.AD*X.BE - kBDoff*X.ABDE
        - kEnz*ds.SunY*X.ABDE;
END ComputeRHS;

PROCEDURE TestZeroStep;
BEGIN
    IF hUsed <= 1/MAX(LONGREAL) THEN
        Out.String('# The simulation terminated with the estimated step = 0');
        Out.Ln;
        Platform.ExitOS(0);
    END
END TestZeroStep;

PROCEDURE Print(composition: RK.State; cycle: LONGINT; time: LONGREAL);
VAR X: Chemicals; Atot, Btot, Dtot, Etot, Stot, Ttot: LONGREAL; i: LONGINT;
BEGIN
    ComputeChemicals(X, composition);
    Stot := X.S + X.DS + X.ES + X.DES + X.ST;
    Ttot := X.T + X.AT + X.BT + X.ABT + X.ST;
    Atot := A0 + S0 - Stot;
    Btot := B0 + S0 - Stot;
    Dtot := D0 + T0 - Ttot;
    Etot := E0 + T0 - Ttot;
    Out.Int(cycle+1,10);Out.String(" ");
    Out.Real(SHORT(time),12);Out.String(" ");
    FOR i := 0 TO 18 DO
        Out.Real(SHORT(composition[i]),11);Out.String(" ");
    END;
    Out.Real(SHORT(Atot),11);Out.String(" ");
    Out.Real(SHORT(Btot),11);Out.String(" ");
    Out.Real(SHORT(Dtot),11);Out.String(" ");
    Out.Real(SHORT(Etot),11);Out.String(" ");
    Out.Real(SHORT(Stot),11);Out.String(" ");
    Out.Real(SHORT(Ttot),11);Out.String(" ");
    Out.Real(SHORT(SunYtot),11);Out.String(" ");
    Out.Real(SHORT(SunY),11);Out.String(" ");
    Out.Real(SHORT(Stot - S0),11);Out.String(" ");
    Out.Real(SHORT(Ttot - T0),11);Out.Ln
END Print;

PROCEDURE PerformReaction;
BEGIN
    time := 0.0;
    temp := temp0 + tcelsius;
    ComputeK(temp);
    elapsedPrintPeriods := 0;
    WHILE time < cycleTime DO
        IF delayModel = Lag THEN IF time < Tlag THEN SunY := 0. ELSE SunY := SunYtot END;
        ELSIF delayModel = Activation THEN SunY := SunYtot*(1 - M.exp(-kact*time));
        ELSE SunY := SunYtot;
        END;
        ds.SunY := SunY;
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriod) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriod);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END PerformReaction;

PROCEDURE KeepHot;
BEGIN
    time := 0.0;
    temp := temp0 + tHeatCelsius;
    ComputeK(temp);
    ds.SunY := 0.0;
    elapsedPrintPeriods := 0;
    WHILE time < hotTime DO
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriodHeat) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriodHeat);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END KeepHot;

PROCEDURE ChangeTemperature(heatingMode: BOOLEAN);
VAR initT, deltaT, targetTime: LONGREAL;
BEGIN
    time := 0.0;
    ds.SunY := 0.0;
    elapsedPrintPeriods := 0;
    IF heatingMode = heating THEN
        initT := temp0 + tcelsius;
        deltaT := tHeatCelsius - tcelsius;
        targetTime := heatingTime;
    ELSE
        initT := temp0 + tHeatCelsius;
        deltaT := tcelsius - tHeatCelsius;
        targetTime := coolingTime;
    END;
    WHILE time < targetTime DO
        temp := initT + deltaT*time/targetTime;
        ComputeK(temp);
        ds.UpdateRKF45(h,hUsed,rtol,atol);
        TestZeroStep;
        time := time + hUsed;
        timeElapsed := timeElapsed + hUsed;
        IF ENTIER(time/printPeriodHeat) > elapsedPrintPeriods THEN
            elapsedPrintPeriods := ENTIER(time/printPeriodHeat);
            Print(ds.state,cycle,timeElapsed);
        END;
    END;
END ChangeTemperature;

PROCEDURE MainLoop;
BEGIN
    NEW(ds); NEW(ds.state,19);
    ds.state[0] := A0; ds.state[1] := B0;
    ds.state[2] := D0; ds.state[3] := E0;
    ds.state[4] := S0; ds.state[5] := T0;
    FOR t := 6 TO 18 DO
        ds.state[t] := 0.0;
    END;
    SunYtot := SunY0; SunY := SunYtot;
    PrintHeadder;
    Print(ds.state,0,0);
    timeElapsed := 0.0;
    h := step;
    heatingTime := (tHeatCelsius - tcelsius)/heatingRamp;
    coolingTime := (tHeatCelsius - tcelsius)/coolingRamp;
    FOR cycle := 0 TO cycleNum - 1 DO
        ChangeTemperature(heating);
        KeepHot;
        ChangeTemperature(cooling);
        IF degSunY THEN SunYtot := SunYtot*M.exp(-kdeg*(hotTime+heatingTime+coolingTime)) END;
        PerformReaction;
    END;
    (*
    temp := temp0 + tcelsius;
    ComputeK(temp);
    PrintK
    *)
END MainLoop;

BEGIN
    IF GetParameters() THEN MainLoop
    ELSE Out.String('Simulation parameters have problems.'); Out.Ln END
END MutualLigationContT.
