#!/bin/python3

import scipy.integrate as spig
import numpy as np
import sys

Rgas = 1.987e-3 # ccal / K mol
temp0 = 273.15 # K

kADon = 0.0; kADoff = 0.0; kATon = 0.0; kAToff = 0.0
kBDon = 0.0; kBDoff = 0.0; kBEon = 0.0; kBEoff = 0.0
kBTon = 0.0; kBToff = 0.0; kSTon = 0.0; kSToff = 0.0
kDSon = 0.0; kDSoff = 0.0; kESon = 0.0; kESoff = 0.0

def GetParameters():
	success = True
	if len(sys.argv) != 64:
		success = False
	else:
		parameters = sys.argv[1:]
	if success:
		for i,entry in enumerate(parameters[:51]):
			parameters[i] = float(entry)
		for i,entry in enumerate(parameters[51:60]):
			parameters[i+51] = int(entry)
		for i,entry in enumerate(parameters[60:62]):
			parameters[i+60] = bool(entry)
		global fCGbd, fCGad, fCGbe, fCGat, \
		fCGbt, fCGds, fCGes, fCGst, \
		Sbd, Sad, Sbe, Sat, \
		Sbt, Sds, Ses, Sst, \
		Hbd, Had, Hbe, Hat, \
		Hbt, Hds, Hes, Hst, \
		A0, B0, E0, D0, \
		T0, S0, SunY0, \
		cycleTime, heatingRamp, \
		hotTime, coolingRamp, \
		step, tcelsius, tHeatCelsius, \
		printPeriod, printPeriodHeat, \
		kdeg, Tlag, kact, kEnz, \
		kUni, kBi, \
		DeltaGCG, DeltaGAU, DeltaGini, \
		rtol, atol, \
		Lbd, Lad, Lbe, Lat, \
		Lbt, Lds, Les, Lst, \
		cycleNum, \
		suppressBDbinding, degSunY, \
		delayModel
		fCGbd, fCGad, fCGbe, fCGat, \
		fCGbt, fCGds, fCGes, fCGst, \
		Sbd, Sad, Sbe, Sat, \
		Sbt, Sds, Ses, Sst, \
		Hbd, Had, Hbe, Hat, \
		Hbt, Hds, Hes, Hst, \
		A0, B0, E0, D0, \
		T0, S0, SunY0, \
		cycleTime, heatingRamp, \
		hotTime, coolingRamp, \
		step, tcelsius, tHeatCelsius, \
		printPeriod, printPeriodHeat, \
		kdeg, Tlag, kact, kEnz, \
		kUni, kBi, \
		DeltaGCG, DeltaGAU, DeltaGini, \
		rtol, atol, \
		Lbd, Lad, Lbe, Lat, \
		Lbt, Lds, Les, Lst, \
		cycleNum, \
		suppressBDbinding, degSunY, \
		delayModel = parameters[:]
	return success

def KD(fCG,temp):
	return np.exp(-2.27*fCG/(0.24 + fCG)/(Rgas*temp))

def Kd(S,H,temp):
	return np.exp((H/temp - S)/Rgas)

def kon(L,fCG,temp):
	return L*kBi*2*kUni/(2*kUni + kBi*KD(fCG,temp))

def ComputeK(temp):
	global kADon, kADoff, kATon, kAToff, \
	kBDon, kBDoff, kBEon, kBEoff, \
	kBTon, kBToff, kSTon, kSToff, \
	kDSon, kDSoff, kESon, kESoff

	kADon  = kon(Lad,fCGad,temp) #(**)
	kADoff = kADon*Kd(Sad,Had,temp)

	kATon  = kon(Lat,fCGat,temp) #(**)
	kAToff = kATon*Kd(Sat,Hat,temp)

	if suppressBDbinding:
		kBDon  = 0.0 #(**)
		kBDoff = 0.0
	else:
		kBDon  = kon(Lbd,fCGbd,temp) #(**)
		kBDoff = kBDon*Kd(Sbd,Hbd,temp)

	kBEon  = kon(Lbe,fCGbe,temp) #(**)
	kBEoff = kBEon*Kd(Sbe,Hbe,temp)

	kBTon  = kon(Lbt,fCGbt,temp) #(**)
	kBToff = kBTon*Kd(Sbt,Hbt,temp)

	kSTon  = kon(Lst,fCGst,temp) #(**)
	kSToff = kSTon*Kd(Sst,Hst,temp)

	kDSon  = kon(Lds,fCGds,temp) #(**)
	kDSoff = kDSon*Kd(Sds,Hds,temp)

	kESon  = kon(Les,fCGes,temp); #(**)
	kESoff = kESon*Kd(Ses,Hes,temp)

def PrintHeader():
	toprint = ['cycle', 't', \
		'[A]', '[B]', '[D]', '[E]', \
		'[S]', '[T]', \
		'[BD]', '[AD]', '[BE]', '[AT]', \
		'[BT]', '[DS]', '[ES]', '[ST]', \
		'[ABT]', '[DES]', '[ABD]', '[BDE]', \
		'[ABDE]', '[Atot]', '[Btot]', '[Dtot]', \
		'[Etot]', '[Stot]', '[Ttot]', \
		'[SunYtot]', '[SunY]', '[Sprod]', '[Tprod]']
	print(toprint,flush=True)

# A    = state[ 0]
# B    = state[ 1]
# D    = state[ 2]
# E    = state[ 3]
# S    = state[ 4]
# T    = state[ 5]
# BD   = state[ 6]
# AD   = state[ 7]
# BE   = state[ 8]
# AT   = state[ 9]
# BT   = state[10]
# DS   = state[11]
# ES   = state[12]
# ST   = state[13]
# ABT  = state[14]
# DES  = state[15]
# ABD  = state[16]
# BDE  = state[17]
# ABDE = state[18]

def RHS(t,y):
	result = np.zeros(len(y))
	result[ 0] = \
		- kADon*y[0]*y[2]   + kADoff*y[7]   - kATon*y[0]*y[5]  + kAToff*y[9] \
		- kADon*y[0]*y[6]  + kADoff*y[16]  - kATon*y[0]*y[10] + kAToff*y[14] \
		- kADon*y[0]*y[17] + kADoff*y[18]
	result[ 1] = \
		- kBDon*y[1]*y[2]  + kBDoff*y[6]  - kBEon*y[1]*y[3]  + kBEoff*y[8] \
		- kBTon*y[1]*y[5]  + kBToff*y[10]  - kBTon*y[1]*y[9] + kBToff*y[14] \
		- kBDon*y[1]*y[7] + kBDoff*y[16]
	result[ 2] = \
		- kADon*y[0]*y[2]  + kADoff*y[7]  - kBDon*y[1]*y[2]  + kBDoff*y[6] \
		- kDSon*y[2]*y[4]  + kDSoff*y[11]  - kDSon*y[2]*y[12] + kDSoff*y[15] \
		- kBDon*y[2]*y[8] + kBDoff*y[17]
	result[ 3] = \
		- kBEon*y[3]*y[1]   + kBEoff*y[8]   - kESon*y[3]*y[4]  + kESoff*y[12] \
		- kBEon*y[3]*y[6]  + kBEoff*y[17]  - kESon*y[3]*y[11] + kESoff*y[15] \
		- kBEon*y[3]*y[16] + kBEoff*y[18]
	result[ 4] = \
		- kSTon*y[4]*y[5] + kSToff*y[13] - kDSon*y[4]*y[2] + kDSoff*y[11] \
		- kESon*y[4]*y[3] + kESoff*y[12]
	result[ 5] = \
		- kSTon*y[5]*y[4] + kSToff*y[13] - kBTon*y[5]*y[1] + kBToff*y[10] \
		- kATon*y[5]*y[0] + kAToff*y[9]
	result[ 6] = \
		- kADon*y[0]*y[6] + kADoff*y[16] + kBDon*y[1]*y[2] - kBDoff*y[6] \
		- kBEon*y[3]*y[6] + kBEoff*y[17]
	result[ 7] = \
		  kADon*y[0]*y[2] - kADoff*y[7] - kBDon*y[1]*y[7] + kBDoff*y[16] \
		- kBDon*y[7]*y[8] + kBDoff*y[18]
	result[ 8] = \
		  kBEon*y[1]*y[3]   - kBEoff*y[8]   - kBDon*y[2]*y[8] + kBDoff*y[17] \
		- kBDon*y[7]*y[8] + kBDoff*y[18]
	result[ 9] = \
		  kATon*y[0]*y[5] - kAToff*y[9] - kBTon*y[1]*y[9] + kBToff*y[14]
	result[10] = \
		- kATon*y[0]*y[10] + kAToff*y[14] + kBTon*y[1]*y[5] - kBToff*y[10] \
		+ kEnz*SunY*y[17]
	result[11] = \
		  kDSon*y[2]*y[4] - kDSoff*y[11] - kESon*y[3]*y[11] + kESoff*y[15] \
		+ kEnz*SunY*y[16]
	result[12] = \
		- kDSon*y[2]*y[12] + kDSoff*y[15] + kESon*y[3]*y[4] - kESoff*y[12]
	result[13] = \
		  kSTon*y[4]*y[5] - kSToff*y[13] \
		+ kEnz*SunY*(y[14] + y[15])
	result[14] = \
		  kATon*y[0]*y[10] - kAToff*y[14] + kBTon*y[1]*y[9] - kBToff*y[14] \
		+ kEnz*SunY*(y[18]/2.0 - y[14])
	result[15] = \
		  kDSon*y[2]*y[12] - kDSoff*y[15] + kESon*y[3]*y[11] - kESoff*y[15] \
		+ kEnz*SunY*(y[18]/2.0 - y[15])
	result[16] = \
		  kADon*y[0]*y[6]  - kADoff*y[16] + kBDon*y[1]*y[7] - kBDoff*y[16] \
		- kBEon*y[3]*y[16] + kBEoff*y[18] \
		- kEnz*SunY*y[16]
	result[17] = \
		- kADon*y[0]*y[17] + kADoff*y[18] + kBDon*y[2]*y[8] - kBDoff*y[17] \
		+ kBEon*y[3]*y[6]  - kBEoff*y[17] \
		- kEnz*SunY*y[17]
	result[18] = \
		  kADon*y[0]*y[17] - kADoff*y[18] + kBEon*y[3]*y[16] - kBEoff*y[18] \
		+ kBDon*y[7]*y[8] - kBDoff*y[18] \
		- kEnz*SunY*y[18]
	return result

def Print(state,cycle,time):
#	Stot := X.S + X.DS + X.ES + X.DES + X.ST
	Stot = state[4] + state[11] + state[12] + \
		state[15] + state[13]
#	Ttot = X.T + X.AT + X.BT + X.ABT + X.ST
	Ttot = state[5] + state[9] + state[10] + \
		state[14] + state[13]
	Atot = A0 + S0 - Stot
	Btot = B0 + S0 - Stot
	Dtot = D0 + T0 - Ttot
	Etot = E0 + T0 - Ttot
	toprint = [cycle+1,time]+\
		list(state)+\
		[Atot,Btot,Dtot,Etot,Stot,Ttot]+\
		[SunYtot,SunY]+\
		[Stot - S0, Ttot - T0]
	print(toprint,flush=True)

def PerformReaction(DS,cycle):
	global timeElapsed, SunY
	time = 0.0
	temp = temp0 + tcelsius
	ComputeK(temp)
	elapsedPrintPeriods = 0
	while time < cycleTime:
		if delayModel == 'Lag':
			if time < Tlag:
				SunY = 0.0
			else:
				SunY = SunYtot
		elif delayModel == 'Activation':
			SunY = SunYtot*(1 - np.exp(-kact*time))
		else:
			SunY = SunYtot
		message = DS.step()
		if message != None:
			print(message)
		hUsed = DS.step_size
		time = time + hUsed
		timeElapsed = timeElapsed + hUsed
		if int(time/printPeriod) > elapsedPrintPeriods:
			elapsedPrintPeriods = int(time/printPeriod)
			Print(DS.y,cycle,timeElapsed)

def KeepHot(DS,cycle):
	global timeElapsed, SunY
	time = 0.0
	temp = temp0 + tHeatCelsius
	ComputeK(temp)
	SunY = 0.0
	elapsedPrintPeriods = 0
	while time < hotTime:
		message = DS.step()
		if message != None:
			print(message)
		hUsed = DS.step_size
		time = time + hUsed
		timeElapsed = timeElapsed + hUsed
		if int(time/printPeriodHeat) > elapsedPrintPeriods:
			elapsedPrintPeriods = int(time/printPeriodHeat)
			Print(DS.y,cycle,timeElapsed)

def ChangeTemperature(mode,DS,cycle):
	global timeElapsed, SunY
	time = 0.0
	SunY = 0.0
	elapsedPrintPeriods = 0
	if mode == 'heating':
		initT = temp0 + tcelsius
		deltaT = tHeatCelsius - tcelsius
		targetTime = heatingTime
	else:
		initT = temp0 + tHeatCelsius
		deltaT = tcelsius - tHeatCelsius
		targetTime = coolingTime
	while time < targetTime:
		temp = initT + deltaT*time/targetTime
		ComputeK(temp)
		message = DS.step()
		if message != None:
			print(message)
		hUsed = DS.step_size
		time = time + hUsed
		timeElapsed = timeElapsed + hUsed
		if int(time/printPeriodHeat) > elapsedPrintPeriods:
			elapsedPrintPeriods = int(time/printPeriodHeat)
			Print(DS.y,cycle,timeElapsed)

def MainLoop():
	init = np.array([A0, B0, D0, E0, S0, T0] + [0.0]*13)
	global SunY, SunYtot
	SunYtot = SunY0
	SunY = SunYtot
	PrintHeader()
	Print(init,0,0.0)
	global timeElapsed, heatingTime, coolingTime
	timeElapsed = 0.0
	heatingTime = (tHeatCelsius - tcelsius)/heatingRamp
	coolingTime = (tHeatCelsius - tcelsius)/coolingRamp
	DS = spig.RK45(RHS,0.0,init,np.inf,first_step=step,rtol=rtol,atol=atol)
	for cycle in range(cycleNum):
		ChangeTemperature('heating',DS,cycle)
		KeepHot(DS,cycle)
		ChangeTemperature('cooling',DS,cycle)
		if degSunY:
			SunYtot = SunYtot*np.exp(-kdeg*(hotTime+heatingTime+coolingTime))
		PerformReaction(DS,cycle)

if GetParameters():
	MainLoop()
else:
	print('Simulation parameters have problems.')
